@@ -2,8 +2,10 @@ from __future__ import annotations
 
 import base64
 import datetime
+import functools
 import hashlib
 import http.client
+import itertools
 import sys
 import threading
 import time
@@ -82,6 +84,36 @@ WEB_ADMIN_SETTINGS_PATH = WEB_ADMIN_BASE_PATH / Path("settings/")
 WEB_ADMIN_MAP_LIST_PATH = WEB_ADMIN_SETTINGS_PATH / Path("maplist/")
 WEB_ADMIN_BANS_PATH = WEB_ADMIN_POLICY_PATH / Path("bans/")
 WEB_ADMIN_SQUADS_PATH = WEB_ADMIN_CURRENT_GAME_PATH / Path("squads/")
+WEB_ADMIN_TRACKING_PATH = WEB_ADMIN_POLICY_PATH / Path("tracking/")
+
+
+def retry(on_exc: Union[Exception, Tuple],
+          tries: int = 10, delay: int = 3, backoff: int = 2,
+          cap: int = 30):
+    """Retry calling the decorated function
+    using an exponential back-off with back-off cap.
+    """
+
+    def deco_retry(f):
+
+        @functools.wraps(f)
+        def f_retry(*args, **kwargs):
+            mtries, mdelay = tries, delay
+            while mtries > 1:
+                try:
+                    return f(*args, **kwargs)
+                except on_exc as e:
+                    logger.info(f"{f.__name__}(): {e}, "
+                                f"retrying in {mdelay} seconds...")
+                    time.sleep(mdelay)
+                    mtries -= 1
+                    mdelay *= backoff
+                    mdelay = min(mdelay, cap)
+            return f(*args, **kwargs)
+
+        return f_retry
+
+    return deco_retry
 
 
 def _in(el: object, seq: Sequence[Sequence]) -> bool:
@@ -223,6 +255,10 @@ class WebAdminAdapter:
             (scheme, netloc, WEB_ADMIN_SQUADS_PATH.as_posix(),
              params, query, fragment)
         )
+        self._tracking_url = urlunparse(
+            (scheme, netloc, WEB_ADMIN_TRACKING_PATH.as_posix(),
+             params, query, fragment)
+        )
 
         # Perform a dummy request to populate self._headers.
         self._perform(self._webadmin_url, skip_auth=True)
@@ -430,7 +466,7 @@ class WebAdminAdapter:
         raise NotImplementedError
 
     def change_map(self, new_map: str, url_extra: dict = None):
-        if new_map.lower() not in [m.lower() for m in self.get_maps()]:
+        if new_map.lower() not in [m.lower() for m in self.get_maps_list()]:
             logger.warning("{nm} not in server map list", nm=new_map)
 
         if url_extra is None:
@@ -486,6 +522,14 @@ class WebAdminAdapter:
 
         return maps
 
+    def get_maps_list(self) -> List[str]:
+        """Return list of all maps of all game modes
+        currently installed on the server.
+        """
+        return list(
+            itertools.chain(*self.get_maps().values())
+        )
+
     def get_players(self) -> List[PlayerWrapper]:
         headers = self._make_auth_headers()
         resp = self._perform(self._players_url, headers=headers)
@@ -506,7 +550,7 @@ class WebAdminAdapter:
             name = player.name
 
         exp_number = ""
-        exp_unit = BAN_EXP_UNIT_PERM = "Never"
+        exp_unit = BAN_EXP_UNIT_PERM
 
         if duration:
             # noinspection PyBroadException
@@ -540,6 +584,9 @@ class WebAdminAdapter:
         # action=revoke&uniqueid=0x01100001013F76D0&playername=&__Submitter=
         raise NotImplementedError
 
+    def revoke_session_ban(self, player: Union[models.Player, PlayerWrapper]):
+        raise NotImplementedError
+
     def get_map_cycles(self) -> List[models.MapCycle]:
         headers = self._make_auth_headers()
         resp = self._perform(self._map_list_url, headers=headers)
@@ -586,6 +633,18 @@ class WebAdminAdapter:
         resp = self._perform(self._squads_url, headers=headers)
         return self._rparser.parse_squads(resp, adapter=self)
 
+    def get_tracked_players(self) -> List[TrackingWrapper]:
+        headers = self._make_auth_headers()
+        resp = self._perform(self._tracking_url, headers=headers)
+        tracked_players = []
+        for _ in range(1):
+            tracked_players.extend(
+                self._rparser.parse_tracking(resp, adapter=self))
+
+        # Remove duplicates in case tracking tab was
+        # mutated during parsing.
+        return list(set(tracked_players))
+
     def _enqueue_chat_messages(self):
         while True and not self._stop_event.is_set():
             self._chat_message_deque.extend(
@@ -601,6 +660,7 @@ class WebAdminAdapter:
         logger.debug("got {clen} chat messages", clen=len(chat_msgs))
         return chat_msgs
 
+    @retry(Exception)
     def _perform(self, url: str, curl_obj: pycurl.Curl = None,
                  headers: dict = None, postfields: dict = None,
                  skip_auth=False) -> bytes:
@@ -639,8 +699,7 @@ class WebAdminAdapter:
             curl_obj.perform()
         except pycurl.error as e:
             logger.debug(e, exc_info=True)
-            logger.warning(e)
-            return b""
+            raise
 
         status = curl_obj.getinfo(pycurl.HTTP_CODE)
         logger.debug("HTTP status: {s}", s=status)
@@ -661,6 +720,12 @@ class WebAdminAdapter:
             raise HTTPError(url=url, msg=phrase,
                             code=status, hdrs=hdrs, fp=None)
 
+        # Server changing maps will trigger sessionid change,
+        # keep track of latest sessionid in response headers.
+        sessionid = self._find_sessionid()
+        if sessionid and self._auth_data:
+            self._auth_data.sessionid = sessionid
+
         return buffer.getvalue()
 
     def _header_function(self, header_line):
@@ -731,12 +796,13 @@ class WebAdminAdapter:
                 logger.error("cannot get sessionid from headers")
                 return r
         except AttributeError as ae:
-            logger.exception(ae)
+            logger.debug(ae, exc_info=True)
             return r
         except Exception as e:
-            logger.exception(e)
+            logger.debug(e, exc_info=True)
             return r
 
+        logger.debug("got sessionid: {si}, from headers", si=r)
         return f'sessionid="{r}";'
 
     def _post_login(self, sessionid: str, token: str,
@@ -784,7 +850,6 @@ class WebAdminAdapter:
             logger.error("unable to get token: {e}", e=ae)
 
         sessionid = self._find_sessionid()
-        logger.debug("got sessionid: {si}, from headers", si=sessionid)
 
         self._post_login(sessionid=sessionid, token=token)
 
@@ -922,6 +987,9 @@ class PlayerWrapper(ModelWrapper):
     def __repr__(self) -> str:
         return f"{self.__class__.__name__}({self.__str__()})"
 
+    def __hash__(self) -> int:
+        return hash(self.player)
+
     @property
     def player(self) -> models.Player:
         assert isinstance(self._model, models.Player)
@@ -963,10 +1031,10 @@ class PlayerWrapper(ModelWrapper):
                                          notify_players)
 
     def revoke_ban(self):
-        raise NotImplementedError
+        self._adapter.revoke_player_ban(self)
 
     def revoke_session_ban(self):
-        raise NotImplementedError
+        pass
 
     def track(self):
         raise NotImplementedError
@@ -1041,10 +1109,62 @@ class BanWrapper(ModelWrapper):
         return self.ban.player
 
     def revoke(self):
-        if isinstance(self._model, models.Ban):
-            self._adapter.revoke_player_ban(self.player)
-        elif isinstance(self._model, models.SessionBan):
-            self._adapter.revoke_session_ban(self.player)
-        else:
-            logger.warn("can't revoke unknown ban type: {t}",
-                        t=type(self))
+        self._adapter.revoke_player_ban(self.player)
+
+
+class SessionBanWrapper(BanWrapper):
+    def __init__(self, ban: models.SessionBan, adapter: WebAdminAdapter):
+        super().__init__(ban, adapter)
+
+    def revoke(self):
+        self._adapter.revoke_session_ban(self.player)
+
+
+class TrackingWrapper(PlayerWrapper):
+
+    def __init__(self, player: models.Player, tracking_data: dict,
+                 adapter: WebAdminAdapter):
+        super().__init__(player, adapter)
+        self._tracking = tracking_data
+
+    def __str__(self) -> str:
+        return f"{self._model.__str__()}, {self._tracking}"
+
+    def __repr__(self) -> str:
+        return f"{self.__class__.__name__}({self.__str__()})"
+
+    @property
+    def tracking(self) -> dict:
+        return self._tracking
+
+    @property
+    def alias(self) -> str:
+        """Current alias."""
+        return self.tracking["Alias"]
+
+    @property
+    def ksb(self) -> str:
+        """Kick, session ban and ban count.
+        A string of format '{K}/{S}/{B}'.
+        E.g. '0/1/1'.
+        """
+        return self.tracking["# K/S/B"]
+
+    @property
+    def num_notes(self) -> int:
+        """Number of notes attached to player."""
+        return self.tracking["# Notes"]
+
+    @property
+    def created(self) -> datetime.datetime:
+        """Creation date of tracking entry."""
+        return self.tracking["Created"]
+
+    @property
+    def num_connects(self) -> datetime.datetime:
+        """The number of times player has connected."""
+        return self.tracking["# Connects"]
+
+    @property
+    def last_seen(self) -> datetime.datetime:
+        return self.tracking["Last Seen"]
